
/*
 *	dratfront.d -- part of the drat! packrat parser package pre-alpha v.0.20
 *
 *	Copyright (C) 2007 Frank F. Hirsch "0ffh"
 *                     (email frank at frankhirsch fullstop net)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be Tful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

module dratfront;

import std.file;
import std.string;
import dratlib;

alias TDratParser!(string) DratParser;
alias TParseNode!(string) PN;
alias TRule!(string) Rule;
// begin user declarations

  alias std.string.toStringz tsz;
  alias string str;

  string theHead=
  "
module MODULENAME;

import std.file;
import std.string;
import dratlib;

alias TDratParser!(RETYPE) DratParser;
alias TParseNode!(RETYPE) PN;
alias TRule!(RETYPE) Rule;
";

  string theParserHead="  DratParser p=new DratParser();\n";

  string theParserTail="  //\n  return p;\n}\n\n";
// end user declarations

TDratParser!(string) generateParser()
{
  DratParser p=new DratParser();
  p.addRule(`top:(ws header notice? body main?)`,
    function string(PN *n)
    {
      PN* header=n.find("header");
      str typestr=header.find("type").find("id").find("expr").match();
      str namestr=header.find("name").find("id").find("expr").match();
      string res;
      if (PN* noticeNode=n.find("notice"))
      {
        string[] lines=noticeNode.findCall("dblock").splitlines();
        foreach (line;lines)
        {
          line=chomp(line);
          res~=line~"\n";
        }
      }
      res~=theHead.replace("MODULENAME",namestr).replace("RETYPE",typestr);
      PN*[] cl=n.find("body").collect("decl");
      foreach (c;cl)
      {
        string[] lines=c.find("dblock").find("inner").match().splitlines();
        res~="// begin user declarations\n";
        foreach (line;lines)
        {
          line=chomp(line);
          res~=line~"\n";
        }
        res~="// end user declarations\n\n";
      }
      res~=`TDratParser!(`~typestr~`) `;
      res~="generateParser()\n";
      res~="{\n";
      res~=theParserHead;
      PN*[] rl=n.find("body").collect("rule");
      foreach (r;rl)
      {
        string rulestr=r.find("decor").call();
        rulestr=rulestr.replace("`","`~"~`"`~"`"~`"`~"~`");
        rulestr="  p.addRule(`"~r.findCall("id")~":"~rulestr~"`";
        res~=rulestr;
        if (PN* blockNode=r.find("dblock"))
        {
          res~=",\n    function "~typestr~"(PN *n)";
          string[] lines=blockNode.match().splitlines();
          foreach (line;lines)
          {
            res~="\n    "~line;
          }
        }
        res~=");\n";
      }
      res~=theParserTail;
      if (PN* mainNode=n.find("main"))
      {
        res~="int main(string[] args)\n{\n";
        res~="  DratParser parser=generateParser();\n";
        string[] lines=mainNode.findCall("dblock").splitlines();
        foreach (line;lines)
        {
          line=chomp(line);
          if (line.length)
            res~="\n"~line;
        }
        res~="\n  return 0;\n}\n";
      }
      return res;
    }
      );
  p.addRule(`header:("parser" ws type:id name:id ";" ws)`);
  p.addRule(`notice:("notice" ws dblock)`);
  p.addRule(`body:(decl / rule / ws)*`);
  p.addRule(`main:("main" ws dblock)`);
  p.addRule(`rule:("rule" ws id ":" ws decor dblock?)`);
  p.addRule(`grammar:(id ":" ws decor ws)`,
    function string(PN *n)
    {
      return n.findCall("id")~":"~n.findCall("decor");
    }
    );
  p.addRule(`decl:("decl" ws dblock)`);
  p.addRule(`dblock:("{" inner:(dws / dstr / dblock / $[^{}"'`~"`"~`/]+$)* "}" ws)`,
    function string(PN *n)
    {
      return n.findMatch("inner");
    }
    );
  p.addRule(`dws:($\s+$ / dlrem / dbrem / dnbrem)`);
  p.addRule(`dlrem:($//[^\r\n]*$)`);
  p.addRule(`dbrem:("/*" $([^*]|(\*[^/]))*$ "*/")`);
  p.addRule(`dnbrem:("/+" ($[^+/]+$ / $\+[^/]$ / $/[^+]$ / dnbrem)* "+/")`);
  p.addRule(`dstr:(string / $`~"`"~`([^`~"`"~`]|(\\`~"`"~`))*`~"`"~`$)`);
  p.addRule(`fst:(ws seq ($[/|]$ ws seq)*)`,
    function string(PN *n)
    {
      string[] l=n.collectCall("seq");
      string s;
      foreach (e;l)
        s~=e~" / ";
      return s[0..$-3];
    }
    );
  p.addRule(`seq:named+`,
    function string(PN *n)
    {
      string[] l=n.collectCall("named");
      string s;
      foreach (e;l)
        s~=e~" ";
      return s[0..$-1];
    }
    );
  p.addRule(`named:((id ":")? ws decor ws)`,
    function string(PN *n)
    {
      PN* s=n.find("id");
      if (s)
        return s.call()~":"~n.findCall("decor");
      else
        return n.findCall("decor");
    }
    );
  p.addRule(`decor:(prefix:$[&!]?$ undec postfix:$[?*+]?$ ws)`,
    function string(PN *n)
    {
      return n.findMatch("prefix")~n.findCall("undec")~n.findMatch("postfix");
    }
    );
  p.addRule(`undec:(atom / "(" ws fst ")" ws)`,
    function string(PN *n)
    {
      PN* s=n.find("atom");
      if (s)
        return s.call();
      else
        return "("~n.findCall("fst")~")";
    }
    );
  p.addRule(`atom:(string / regexp / id)`,
    function string(PN *n)
    {
      return n.l.call();
    }
    );
  p.addRule(`string:(expr:($'([^']|(\\'))*'$ / %"([^"]|(\\"))*"%) ws)`,
    function string(PN *n)
    {
      return n.find("expr").match();
    }
    );
  p.addRule(`regexp:(expr:(("$" %([^$]|(\$\$))*% "$") / ("%" $([^%]|(\%\%))*$ "%")) ws)`,
    function string(PN *n)
    {
      return n.find("expr").match();
    }
    );
  p.addRule(`id:(expr:($[_a-zA-Z][-_a-zA-Z0-9]*$) ws)`,
    function string(PN *n)
    {
      return n.find("expr").match();
    }
    );
  p.addRule(`ws:($\s+$ / $#[^\r\n]*$)*`);
  //
  return p;
}

int main(string[] args)
{
  DratParser parser=generateParser();

  if (args.length<2)
  {
    printf("no file name specified\n");
    return 1;
  }
  string src=cast(string)std.file.read(args[1]~".drat");
  if (PN *r=parser.parse("top",src))
  {
    std.file.write(args[1]~".d",cast(void[])r.call());
  }
  else
  {
    int l,o;
    parser.idx2lno(parser.src,parser.spmax,&l,&o);
    printf("fail at line %i offset %i\n",l,o);
  }
  return 0;
}
